import importlib
import pkgutil
from inspect import isabstract
from typing import Type, Callable, Union


def to_id(value: Union[Type, Callable]):
    return value.__module__ + "." + value.__name__


def import_submodules(package, recursive=True):
    """ Import all submodules of a module, recursively, including subpackages

    :param package: package (name or actual module)
    :type package: str | module
    :rtype: dict[str, types.ModuleType]
    """
    if isinstance(package, str):
        package = importlib.import_module(package)
    results = {}
    for loader, name, is_pkg in pkgutil.walk_packages(package.__path__):
        full_name = package.__name__ + '.' + name
        try:
            results[full_name] = importlib.import_module(full_name)
        except ModuleNotFoundError:
            continue
        if recursive and is_pkg:
            results.update(import_submodules(full_name))
    return results


def get_all_subclasses(cls):
    all_subclasses = set()

    for subclass in cls.__subclasses__():
        all_subclasses.add(subclass)
        all_subclasses = all_subclasses.union(get_all_subclasses(subclass))

    return all_subclasses


def get_all_non_abstract_subclasses(cls):
    import_submodules(str(cls.__module__).split('.')[0])
    all_subclasses = get_all_subclasses(cls)
    return [s for s in all_subclasses if not isabstract(s)]



import random

import numpy as np
import pandas as pd
from sklearn import preprocessing
from sklearn.metrics import roc_auc_score, accuracy_score, precision_score, recall_score, f1_score, \
    matthews_corrcoef as sk_matthews_corrcoef


def seed_everything(seed=1):
    """
    Sets up seed for random and numpy random.

    :param seed: random seed

    """
    random.seed(seed)
    np.random.seed(seed)


def auc_score(y_true: np.ndarray, y_pred: np.ndarray, average=None, multi_class=None,
              labels: np.ndarray = None) -> float:
    """
    Calculates multiclass AUC score.

    :param y_true: real labels
    :param y_pred: predicted labels
    :param average: sklearn roc_auc_score param
    :param multi_class: sklearn roc_auc_score param
    :param labels: sklearn roc_auc_score param
    :return: AUC value

    """
    lb = preprocessing.LabelBinarizer()
    lb.fit(labels)
    y_score = lb.transform(y_pred)
    return roc_auc_score(y_true, y_score, average=average, multi_class=multi_class, labels=labels)


def to_latex(df: pd.DataFrame, **kwargs) -> str:
    """
    Converts Pandas DateFrame to latex table string.

    :param df: Pandas DateFrame with data to be converted
    :param kwargs: other pandas df.to_latex args
    :return: Latex string

    """
    tex = df.style.to_latex(**kwargs)
    tex = tex.replace('_', ' ').replace('\\toprule\n', '').replace('\\midrule\n', '').replace('\\bottomrule\n', '')
    hline = '\\hline'
    lines = tex.split('\n')
    lines.insert(1, '\\footnotesize')
    lines.insert(5, hline)
    lines.insert(7, hline)
    lines.insert(-3, hline)

    # this fixes difference between python 3.7 and later versions
    lines[6] = lines[6].replace('{', '').replace('}', '')
    return '\n'.join(lines)


def accuracy(y_true, y_pred, labels):
    return accuracy_score(y_true, y_pred)


def precision(y_true, y_pred, labels):
    return precision_score(y_true, y_pred, average='weighted', labels=labels,
                           zero_division=0)


def recall(y_true, y_pred, labels):
    return recall_score(y_true, y_pred, average='weighted', labels=labels,
                        zero_division=0)


def f1(y_true, y_pred, labels):
    return f1_score(y_true, y_pred, average='weighted', labels=labels, zero_division=0)


def auc(y_true, y_pred, labels):
    return auc_score(y_true, y_pred, average='weighted', multi_class='ovo', labels=labels)


def matthews_corrcoef(y_true, y_pred, labels):
    return sk_matthews_corrcoef(y_true, y_pred)



import dash
import dash_ace
import dash_bootstrap_components as dbc
from dash import html, Output, callback, Input, State, no_update
from dash.dcc import Store

from cacp.gui.components.shared.utils import global_location_href_output
from cacp.gui.db.custom_metrics import get_custom_metric, CustomMetricType, CustomMetric, \
    update_custom_metric, CUSTOM_METRIC_BATCH_CODE_TEMPLATE, CUSTOM_INCREMENTAL_METRIC_CODE_TEMPLATE, \
    test_custom_metric_code
from cacp.gui.preview import preview_button_kwargs

dash.register_page(__name__, path_template="/custom_metric/<custom_metric_id>")

CUSTOM_METRIC_NAME_ID = "custom_metric_name_id"
CUSTOM_METRIC_TYPE_ID = "custom_metric_type_id"
CUSTOM_METRIC_CODE_ID = "custom_metric_code_id"
CUSTOM_METRIC_SAVE_BUTTON_ID = "custom_metric_save_button_id"
CUSTOM_METRIC_TOAST_ID = "custom_metric_toast_id"
CUSTOM_METRIC_STORE_ID = "custom_metric_store_id"


def layout(custom_metric_id=None):
    custom_metric = get_custom_metric(custom_metric_id) if custom_metric_id else {}

    name_input = html.Div(
        [
            dbc.Label("Name", html_for=CUSTOM_METRIC_NAME_ID),
            dbc.Input(id=CUSTOM_METRIC_NAME_ID, placeholder="Enter name", value=custom_metric.get("name")),
            dbc.FormText(
                "Provide name for your metric, it should be unique.",
                color="secondary",
            ),
        ],
        className="mb-3",
    )

    type_input = html.Div(
        [
            dbc.Label("Type", html_for=CUSTOM_METRIC_TYPE_ID),
            dbc.RadioItems(
                {
                    CustomMetricType.BATCH: "Batch learning",
                    CustomMetricType.INCREMENTAL: "Incremental learning"
                },
                value=custom_metric.get("type"),
                id=CUSTOM_METRIC_TYPE_ID
            )
        ],
        className="py-2",
    )

    code = html.Div(
        [
            dbc.Label("Code", html_for=CUSTOM_METRIC_NAME_ID), dash_ace.DashAceEditor(
            id=CUSTOM_METRIC_CODE_ID,
            value=custom_metric.get("code"),
            theme='github',
            mode='python',
            tabSize=2,
            enableBasicAutocompletion=True,
            placeholder='Python code ...',
            style={"height": "calc(100vh - 400px)", "width": "100%"}
        )])

    save_button = html.Div([
        dbc.Button(
            "Save", id=CUSTOM_METRIC_SAVE_BUTTON_ID, className="mt-3", n_clicks=0, **preview_button_kwargs()
        )
    ], className="d-flex justify-content-end align-items-center")

    toast = dbc.Toast(
        "",
        id=CUSTOM_METRIC_TOAST_ID,
        header="Validation failed",
        is_open=False,
        dismissable=True,
        duration=5000,
        style={"position": "fixed", "top": 66, "right": 10, "width": 350, "z-index": "2000"},
    )

    store = Store(id=CUSTOM_METRIC_STORE_ID, data=custom_metric)

    view = html.Div([
        name_input,
        type_input,
        code,
        save_button,
        toast,
        store
    ])

    return view


@callback(
    Output(CUSTOM_METRIC_CODE_ID, "value"),
    Input(CUSTOM_METRIC_TYPE_ID, "value"),
    State(CUSTOM_METRIC_STORE_ID, "data"),
    prevent_initial_call=True,
)
def update_river_metrics_button_class(type_value: CustomMetricType, metric: CustomMetric):
    if type_value == CustomMetricType.BATCH:
        return CUSTOM_METRIC_BATCH_CODE_TEMPLATE.format(metric.get("id"))
    if type_value == CustomMetricType.INCREMENTAL:
        return CUSTOM_INCREMENTAL_METRIC_CODE_TEMPLATE.format(metric.get("id"))
    return no_update


@callback(
    global_location_href_output(),
    Output(CUSTOM_METRIC_TOAST_ID, "children"),
    Output(CUSTOM_METRIC_TOAST_ID, "header"),
    Output(CUSTOM_METRIC_TOAST_ID, "is_open"),
    Input(CUSTOM_METRIC_SAVE_BUTTON_ID, "n_clicks"),
    State(CUSTOM_METRIC_NAME_ID, "value"),
    State(CUSTOM_METRIC_TYPE_ID, "value"),
    State(CUSTOM_METRIC_CODE_ID, "value"),
    State(CUSTOM_METRIC_STORE_ID, "data"),
    prevent_initial_call=True,
)
def save_button_click(
    n_clicks, name_value: str, type_value: CustomMetricType, code_value: str, custom_metric: CustomMetric
):
    href, toast_message, toast_header, toast_is_open, = no_update, "", no_update, False
    if n_clicks:
        custom_metric_id = int(custom_metric["id"])
        errors = []
        if not name_value:
            errors.append("Name can not be empty.")

        test_error = test_custom_metric_code(custom_metric_id, code_value, type_value)
        if test_error:
            errors.append(test_error)

        if errors:
            toast_message = dbc.Alert([html.Div(e) for e in errors], color="danger")
            toast_header = "Validation error"
        else:
            update_custom_metric(custom_metric_id, name_value, type_value, code_value)
            toast_message = dbc.Alert(html.Div("Saved"))
            toast_header = "Success"
        toast_is_open = True
    return href, toast_message, toast_header, toast_is_open
